# Copyright (c) 2022 International Institute for Applied Systems Analysis.
# All rights reserved. The use and distribution terms for this software
# are covered by the MIT License (http://opensource.org/licenses/MIT)
# which can be found in the file LICENSE at the root of this distribution.
# By using this software in any fashion, you are agreeing to be bound by
# the terms of this license.
# You must not remove this notice, or any other, from this software.
"""Execution command checker."""

__author__ = "Anna Shchiptsova"
__copyright__ = "Copyright (c) 2022 IIASA"
__license__ = "MIT"

from getopt import getopt, GetoptError


def _fail(arg, label, vals):
    """Terminate execution with invalid arguments error message."""
    print(
        'Failed to validate: ' +
        arg +
        ' not supported ' +
        label +
        '\n(supported ' +
        ' '.join(map(str, vals)) +
        ')')


def parse(argv, options):
    """Parse command line arguments."""
    help_msg = 'Usage: dice-climate-simulator' + \
        ' -e <emissions> -r <ratio> -f <folder>'

    try:
        opts, _ = getopt(
            argv,
            'h:e:r:f:',
            ['emissions=', 'ratio=', 'folder='])
    except GetoptError:
        print(help_msg)
        return None

    args = {}
    args['folder'] = None
    for opt, arg in opts:
        if opt == '-h':
            print(help_msg)
            return None
        if opt in ('-e', '--emissions'):
            args['emissions'] = arg
        if opt in ('-r', '--ratio'):
            args['ratio'] = arg
        if opt in ('-f', '--folder'):
            args['folder'] = arg

    if 'emissions' not in args:
        print(help_msg)
        return None

    if args['emissions'] == 'hansel2020':
        args['ratio'] = 'hansel2020'

    if args['emissions'] not in options['emissions']:
        _fail(args['emissions'], 'emissions', options['emissions'])
    elif (args['emissions'] != 'hansel2020') and \
        ('ratio' in args) and \
            (args['ratio'] not in options['ratio']):
        _fail(args['ratio'], 'ratio', options['ratio'])
    elif ('ratio' not in args) or \
            ((args['emissions'] != 'hansel2020') and (args['folder'] is None)):
        print(help_msg)
    else:
        return args

    return None



# Copyright (c) 2022 International Institute for Applied Systems Analysis.
# All rights reserved. The use and distribution terms for this software
# are covered by the MIT License (http://opensource.org/licenses/MIT)
# which can be found in the file LICENSE at the root of this distribution.
# By using this software in any fashion, you are agreeing to be bound by
# the terms of this license.
# You must not remove this notice, or any other, from this software.
"""Python climate module implementation from Hansel et al. (2020).

References
----------
.. [1] Hansel, M., Drupp, M., Johansson, D., Nesje, F., Azar, C., Freeman, M.,
    Groom, B., & Sterner, T. (2020). Climate Economics Support for the UN
    Climate Targets. Nature Climate Change, 10: 781-789.
    https://doi.org/10.1038/s41558-020-0833-x
.. [2] Smith, C. J., Forster, P. M.,  Allen, M., Leach, N., Millar, R. J.,
    Passerello, G. A., & Regayre, L. A. (2018). FAIR v1.3: a Simple Emissions-
    Based Impulse Response and Carbon Cycle Model. Geoscientific Model
    Development, 11: 2273–2297. https://doi.org/10.5194/gmd-11-2273-2018
.. [3] Geoffroy, O., Saint-Martin, D., Olivié, D. J. L., Voldoire, A.,
    Bellon, G., & Tytéca, S. (2013). Transient Climate Response in a Two-Layer
    Energy-Balance Model. Part I: Analytical Solution and Parameter Calibration
    Using CMIP5 AOGCM Experiments. Journal of Climate, 26: 1841–1857.
    https://doi.org/10.1175/JCLI-D-12-00195.1

"""

__author__ = "Anna Shchiptsova"
__copyright__ = "Copyright (c) 2022 IIASA"
__license__ = "MIT"



# Copyright (c) 2022 International Institute for Applied Systems Analysis.
# All rights reserved. The use and distribution terms for this software
# are covered by the MIT License (http://opensource.org/licenses/MIT)
# which can be found in the file LICENSE at the root of this distribution.
# By using this software in any fashion, you are agreeing to be bound by
# the terms of this license.
# You must not remove this notice, or any other, from this software.
"""Forward and backward processing of I/O data."""

__author__ = "Anna Shchiptsova"
__copyright__ = "Copyright (c) 2022 IIASA"
__license__ = "MIT"

import os
import csv
from importlib.resources import open_text
import numpy as np  # type: ignore
from climate_module import resources


def _parse_emissions(file, transform, n_pars):
    """Parse net emissions curve values from csv file."""
    csvreader = csv.reader(file)
    header = next(csvreader)

    ids = []
    values = []
    for row in csvreader:
        ids.append(row[:n_pars])
        values.append(transform(np.array(row[n_pars:], dtype=float)))

    return header, ids, values


def read_emissions(
    resource,
    in_resources=False,
    transform=lambda x: x,
        n_pars=6):
    """Read net emissions pathways from csv file."""
    if in_resources:
        with open_text(resources, resource) as file:
            content = _parse_emissions(file, transform, n_pars)
    else:
        content = None
        try:
            with open(resource, 'r', encoding='utf8') as file:
                content = _parse_emissions(file, transform, n_pars)
        except IOError as error:
            print(f'open failed: unable to read file {resource}:\n{error}')

    return content


def read_other_rf_ratio(resource, ssp, ratio):
    """Read radiative forcing from non-CO2 curves from csv resources file."""
    file = open_text(resources, resource)
    csvreader = csv.reader(file)

    for row in csvreader:
        if row[0] == ssp and row[1] == ratio:
            other_rf = np.array(row[2:], dtype=float)
            break

    return other_rf


def write_output(args, variable, header, par_ids, data):
    """Write temperature pathways into csv file."""
    folder = os.path.join('climate', 'scenario ' + args['emissions'])
    if not os.path.exists(folder):
        os.makedirs(folder)

    filename = variable + ' (other_rf ' + args['ratio'] + ').csv'
    dir_path = None

    try:
        with open(
            os.path.join(folder, filename),
            'w',
            newline='',
                encoding='utf8') as file:

            csvwriter = csv.writer(file)
            csvwriter.writerow(header)

            for par_id, values in zip(par_ids, data):
                out = list(map(lambda x: format(x, '.3f'), values))
                csvwriter.writerow([*par_id, *out])

        dir_path = os.path.abspath(folder)

    except IOError as error:
        print(f'save failed: unable to write to {dir_path}:\n{error}')

    return dir_path



# Copyright (c) 2022 International Institute for Applied Systems Analysis.
# All rights reserved. The use and distribution terms for this software
# are covered by the MIT License (http://opensource.org/licenses/MIT)
# which can be found in the file LICENSE at the root of this distribution.
# By using this software in any fashion, you are agreeing to be bound by
# the terms of this license.
# You must not remove this notice, or any other, from this software.
"""Climate module parameters.

Climate module parameters include initial values (Hansel et al. 2020;
DICE2016), parameters of carbon cycle (Smith et al. 2018; Hansel et al. 2020)
and parameters of energy balance model (Hansel et al. 2020).
"""

__author__ = "Anna Shchiptsova"
__copyright__ = "Copyright (c) 2022 IIASA"
__license__ = "MIT"

import numpy as np  # type: ignore


#: General:
#: years per period.
DT = 5

#: Initial cumulative CO2 emissions (GtCO2).
ECUM0 = 400 + 197

#: Initial atmospheric temperature change (°C from 1900);
#: adjusted to only include athropogenic forcing.
TAT0 = 1.243

#: Initial lower stratum temperature change (°C from 1900);
#: adjusted to only include athropogenic forcing.
TLO0 = 0.324

#: Carbon cycle:
#: pre-industrial time-integrated airborne fraction of CO2 (yr).
R0 = 35.0

#: Carbon cycle:
#: change in time-integrated airborne fraction with CO2 (yr/GtC).
RC = 0.019

#: Carbon cycle:
#: change in time-integrated airborne fraction with temperature
#: (yr/K).
RT = 4.165

#: Carbon cycle:
#: partition coefficient of carbon boxes.
A = np.array([0.217, 0.224, 0.282, 0.276])

#: Carbon cycle:
#: present-day decay time constants of CO2 (yr).
TAU = np.array([1000000, 394.4, 36.54, 4.304])

#: Carbon cycle:
#: box parameters.
CARBON_BOXES0 = np.array([127.159, 93.313, 37.840, 7.721])

#: Carbon cycle:
#: equilibrium concentration in atmosphere (GtC).
C_CO2_EQ = 588

#: Energy balance model:
#: equilibrium temperature impact (°C per doubling C02).
NU = 3.1

#: Energy balance model:
#: forcings of equilibrium CO2 doubling (Wm-2)
KAPPA = 3.6813

#: Energy balance model:
#: speed of adjustment parameter for atmospheric temperature.
XI1 = 7.3

#: Energy balance model:
#: climate model parameter.
XI2 = KAPPA / NU

#: Energy balance model:
#: coefficient of heat loss from atmosphere to oceans.
XI3 = 0.73

#: Energy balance model:
#: coefficient of heat gain by deep oceans.
XI4 = 106



# Copyright (c) 2022 International Institute for Applied Systems Analysis.
# All rights reserved. The use and distribution terms for this software
# are covered by the MIT License (http://opensource.org/licenses/MIT)
# which can be found in the file LICENSE at the root of this distribution.
# By using this software in any fashion, you are agreeing to be bound by
# the terms of this license.
# You must not remove this notice, or any other, from this software.
"""Run package without build."""

__author__ = "Anna Shchiptsova"
__copyright__ = "Copyright (c) 2022 IIASA"
__license__ = "MIT"

import sys
from .core import simulate_temperature

if __name__ == '__main__':
    sys.exit(simulate_temperature(sys.argv[1:]))



# Copyright (c) 2022 International Institute for Applied Systems Analysis.
# All rights reserved. The use and distribution terms for this software
# are covered by the MIT License (http://opensource.org/licenses/MIT)
# which can be found in the file LICENSE at the root of this distribution.
# By using this software in any fashion, you are agreeing to be bound by
# the terms of this license.
# You must not remove this notice, or any other, from this software.
"""Climate module equations from Hansel et al. (2020).

Climate module includes:

1) carbon cycle representation from the FAIR climate model (Smith et al. 2018),

2) energy balance model based on Geoffroy (2013) with dynamic non-CO2 forcings.
"""

__author__ = "Anna Shchiptsova"
__copyright__ = "Copyright (c) 2022 IIASA"
__license__ = "MIT"

from functools import reduce
import numpy as np                  # type: ignore
from scipy.optimize import brentq   # type: ignore
from . import parameters as pars


def _co2_to_c(co2):
    """Convert GtCO2 to GtC."""
    return co2 / 3.666


def _init_emissions(e_co2):
    """Calculate cumulative net CO2 emissions curve (GtC)."""
    ecum_c = np.zeros(e_co2.size)
    ecum_c[0] = pars.ECUM0
    for idx in range(1, e_co2.size):
        ecum_c[idx] = ecum_c[idx - 1] + _co2_to_c(e_co2[idx - 1]) * pars.DT

    return ecum_c


def _iirf_equation(alpha, ecum_c, c_co2, temp_atm):
    """Equation of the 100-year average airborne fraction of a pulse of CO2."""
    iirf1 = pars.R0 + pars.RC * (ecum_c - (c_co2 - pars.C_CO2_EQ)) + \
        pars.RT * temp_atm
    iirf2 = alpha * (np.sum(pars.A * pars.TAU *
                            (1 - np.exp(-100 / (pars.TAU * alpha)))))

    return iirf2 - iirf1


def _carbon_concentration(carbon_boxes):
    """Calculate carbon concentration in atmosphere (GtC)."""
    return np.sum(carbon_boxes) + pars.C_CO2_EQ


def _carbon_cycle(e_co2, ecum_c, c_co2, temp_atm, carbon_boxes):
    """Carbon cycle model."""
    alpha = brentq(_iirf_equation, 0.01, 100, args=(ecum_c, c_co2, temp_atm))

    carbon_boxes_next = np.zeros(carbon_boxes.size)
    for idx1 in range(carbon_boxes.size):
        step = 0
        for idx2 in range(5):
            step += np.exp(- (pars.DT - idx2) / (alpha * pars.TAU[idx1]))

        carbon_boxes_next[idx1] = carbon_boxes[idx1] * \
            np.exp(-pars.DT / (alpha * pars.TAU[idx1])) + \
            pars.A[idx1] * _co2_to_c(e_co2) * step

    c_co2_next = _carbon_concentration(carbon_boxes_next)

    return c_co2_next, carbon_boxes_next, alpha


def _energy_balance_model(
    c_co2,
    other_rf,
    absolute_other_rf,
    temp_atm,
        temp_lo):
    """Energy balance model with dynamic non-CO2 forcings."""
    co2_rf = (pars.KAPPA / np.log(2.)) * np.log(c_co2 / pars.C_CO2_EQ)

    if absolute_other_rf:
        total_rf = co2_rf + other_rf
    else:
        total_rf = (1 + other_rf) * co2_rf

    temp_atm_next = reduce(
        lambda y, idx: y + 1 / pars.XI1 * (total_rf - pars.XI2 * y - pars.XI3 *
                                           (y - temp_lo)),
        range(4),
        temp_atm)

    temp_lo_next = temp_lo + pars.DT * pars.XI3 / pars.XI4 * \
        (temp_atm - temp_lo)

    return temp_atm_next, temp_lo_next


def climate_module(e_co2, other_rf, absolute_other_rf):
    """Temperature and carbon concentration pathways from Hansel et al. (2020).

    Climate module can be run either with given absolute non-CO2 forcings or
    with given ratio of non-CO2 to CO2 forcings.
    """
    ecum_c = _init_emissions(e_co2)
    c_co2 = np.zeros(e_co2.size)
    temp_atm = np.zeros(e_co2.size)

    alpha = np.zeros(e_co2.size - 1)

    carbon_boxes = pars.CARBON_BOXES0
    c_co2[0] = _carbon_concentration(carbon_boxes)
    temp_atm[0] = pars.TAT0
    temp_lo = pars.TLO0
    for idx in range(e_co2.size - 1):
        c_co2[idx + 1], carbon_boxes, alpha[idx] = _carbon_cycle(
            e_co2[idx],
            ecum_c[idx],
            c_co2[idx],
            temp_atm[idx],
            carbon_boxes)
        temp_atm[idx + 1], temp_lo = _energy_balance_model(
            c_co2[idx + 1],
            other_rf[idx + 1],
            absolute_other_rf,
            temp_atm[idx],
            temp_lo)

    return temp_atm, c_co2



# Copyright (c) 2022 International Institute for Applied Systems Analysis.
# All rights reserved. The use and distribution terms for this software
# are covered by the MIT License (http://opensource.org/licenses/MIT)
# which can be found in the file LICENSE at the root of this distribution.
# By using this software in any fashion, you are agreeing to be bound by
# the terms of this license.
# You must not remove this notice, or any other, from this software.
"""Command line operations for climate module.

References
----------
.. [1] Hansel, M., Drupp, M., Johansson, D., Nesje, F., Azar, C., Freeman, M.,
    Groom, B., & Sterner, T. (2020). Climate Economics Support for the UN
    Climate Targets. Nature Climate Change, 10: 781-789.
    https://doi.org/10.1038/s41558-020-0833-x
.. [2] [SSP Database] https://tntcat.iiasa.ac.at/SspDb/

"""

__author__ = "Anna Shchiptsova"
__copyright__ = "Copyright (c) 2022 IIASA"
__license__ = "MIT"

import sys
from .core import simulate_temperature


def main():
    """Entry point for the application script."""
    sys.exit(simulate_temperature(sys.argv[1:]))



# Copyright (c) 2022 International Institute for Applied Systems Analysis.
# All rights reserved. The use and distribution terms for this software
# are covered by the MIT License (http://opensource.org/licenses/MIT)
# which can be found in the file LICENSE at the root of this distribution.
# By using this software in any fashion, you are agreeing to be bound by
# the terms of this license.
# You must not remove this notice, or any other, from this software.
"""Climate module (Hansel et al.) for DICE-like simulation model.

References
----------
.. [1] Hansel, M., Drupp, M., Johansson, D., Nesje, F., Azar, C., Freeman, M.,
    Groom, B., & Sterner, T. (2020). Climate Economics Support for the UN
    Climate Targets. Nature Climate Change, 10: 781-789.
    https://doi.org/10.1038/s41558-020-0833-x
.. [2] [SSP Database] https://tntcat.iiasa.ac.at/SspDb/

"""

__author__ = "Anna Shchiptsova"
__copyright__ = "Copyright (c) 2022 IIASA"
__license__ = "MIT"

from climate_module.hansel2020.inputs import (
    radiative_forcing_non_co2,
    emissions_land_use_co2)
from climate_module.hansel2020.model import climate_module
from .options import parse
from .provider import read_emissions, read_other_rf_ratio, write_output


def _options():
    """Supported command line options."""
    return {
        'emissions': ['hansel2020', 'SSP1', 'SSP2', 'SSP3', 'SSP4', 'SSP5'],
        'ratio': ['avg', 'max', 'min']
    }


def _emissions(label, folder):
    """Map emissions curve data by label."""
    if label == 'hansel2020':
        return read_emissions(
            'hansel FFI.csv',
            in_resources=True,
            transform=lambda e: e + emissions_land_use_co2(e.size))

    return read_emissions(folder + '/net-emissions ' + label + '.csv')


def _other_rf(label, ratio):
    """Map non-CO2 forcings data by label."""
    if ratio == 'hansel2020':
        return radiative_forcing_non_co2, True

    labels = {'avg': 'Avg Ratio', 'max': 'Max', 'min': 'Min'}
    return read_other_rf_ratio(
        'SSP markers ratio non_CO2 to CO2.csv',
        label,
        labels[ratio]), False


def simulate_temperature(argv) -> int:
    """Perform climate simulations from generic DICE model."""
    args = parse(argv, _options())

    if args is None:
        return 1

    content = _emissions(
        args['emissions'],
        args['folder'])

    if content is None:
        return 1

    header, par_ids, net_emissions_paths = content

    other_rf, absolute_other_rf = _other_rf(
        args['emissions'],
        args['ratio'])

    climate_output = [[], []]  # type: ignore
    for net_emissions in net_emissions_paths:
        temp_atm, concentration = climate_module(
            net_emissions,
            other_rf,
            absolute_other_rf)
        climate_output[0].append(temp_atm)
        climate_output[1].append(concentration)

    for variable, data in zip(
        ['temperature change', 'concentration'],
            climate_output):
        folder = write_output(args, variable, header, par_ids, data)

    if folder is None:
        return 1

    print('Results saved to: ' + folder)
    return 0



# Copyright (c) 2022 International Institute for Applied Systems Analysis.
# All rights reserved. The use and distribution terms for this software
# are covered by the MIT License (http://opensource.org/licenses/MIT)
# which can be found in the file LICENSE at the root of this distribution.
# By using this software in any fashion, you are agreeing to be bound by
# the terms of this license.
# You must not remove this notice, or any other, from this software.
"""Net CO2 emissions and radiative forcing curves from Hansel et al. (2020)."""

__author__ = "Anna Shchiptsova"
__copyright__ = "Copyright (c) 2022 IIASA"
__license__ = "MIT"

import numpy as np  # type: ignore

#: Net CO2 FFI emissions curve (GtCO2).
emissions_ffi_co2 = np.array([
    35.7404, 37.6787, 29.0282, 29.4648, 29.4406,
    28.9376, 27.9357, 26.4317, 24.4262, 21.931,
    18.9743, 15.5854, 11.7991, 7.6487, 3.16449,
    -1.62853, -6.71595, -12.1001])

#: Curve with ratios of non-CO2 radiative forcing to CO2 radiative forcing (-).
radiative_forcing_non_co2 = np.array([
    0.31, 0.455, 0.514, 0.58, 0.586,
    0.592, 0.562, 0.533, 0.496, 0.462,
    0.443, 0.425, 0.411, 0.397, 0.382,
    0.367, 0.352, 0.337])


def emissions_land_use_co2(n_points):
    """AFOLU emissions (GtCO2)."""
    e_points = np.zeros(n_points)
    e_points[0] = 2.6

    for x_points in range(1, n_points):
        e_points[x_points] = e_points[x_points - 1] * (1 - 0.115)

    return e_points
